==> cursors_extended-fabric-4.1.0%2B1.21.9-sources.jar: io/github/fishstiz/cursors_extended/cursor/debug/CursorDebugRendererImpl.java <==
package io.github.fishstiz.cursors_extended.cursor.debug;

import io.github.fishstiz.cursors_extended.CursorsExtended;
import io.github.fishstiz.cursors_extended.platform.Services;
import io.github.fishstiz.cursors_extended.util.CursorTypeUtil;
import io.github.fishstiz.cursors_extended.util.DrawUtil;
import org.jetbrains.annotations.Nullable;
import org.joml.Matrix3x2fStack;

import java.util.function.Supplier;
import net.minecraft.class_124;
import net.minecraft.class_2561;
import net.minecraft.class_310;
import net.minecraft.class_332;
import net.minecraft.class_364;
import net.minecraft.class_4069;
import net.minecraft.class_437;
import net.minecraft.class_5244;
import net.minecraft.class_5251;
import net.minecraft.class_8021;
import net.minecraft.class_8030;

class CursorDebugRendererImpl implements CursorDebugRenderer {
    private static final float TEXT_SCALE = 0.75f;
    private final class_2561 screenLabel = class_2561.method_43470("S: ").method_54663(0xFF339BFF); // blue
    private final class_2561 deepestLabel = class_2561.method_43470("D: ").method_54663(0xFF00FF00); // green
    private final class_2561 inspectedLabel = class_2561.method_43470("I: ").method_54663(0xFFFF0000); // red
    private final class_2561 virtualModeLabel = class_2561.method_43470("Virtual Mode: ").method_27692(class_124.field_1065);
    private class_8030 inspectedBounds;
    private String inspectedElement;

    @Override
    public boolean isActive() {
        return true;
    }

    @Override
    public void setLastCursorAt(class_364 inspected, double mouseX, double mouseY) {
        if (CursorTypeUtil.isHovered(inspected, mouseX, mouseY)) {
            this.inspectedBounds = getBounds(inspected);
            this.inspectedElement = getClassName(inspected);
        }
    }

    @Override
    public void render(class_310 minecraft, Supplier<@Nullable class_437> visibleScreen, class_332 guiGraphics, double mouseX, double mouseY) {
        class_437 screen = visibleScreen.get();
        if (screen != null) {
            class_8030 screenRectangle = getBounds(screen);
            renderScreenName(minecraft, screen, screenRectangle, guiGraphics);
            renderInspected(minecraft, renderDeepest(minecraft, screen, guiGraphics, mouseX, mouseY), guiGraphics);
            renderVirtualInfo(minecraft, screenRectangle, guiGraphics);
        }
    }

    private class_8030 renderDeepest(class_310 minecraft, class_437 screen, class_332 guiGraphics, double mouseX, double mouseY) {
        class_364 child = findDeepest(screen, mouseX, mouseY);
        class_364 inspect = child != null ? child : screen;
        class_8030 bounds = getBounds(inspect);
        class_2561 label = deepestLabel.method_27661().method_27693(getClassName(inspect));
        renderInfo(minecraft, guiGraphics, bounds, label, Position.TOP_LEFT, true);
        return bounds;
    }

    private void renderInspected(class_310 minecraft, class_8030 container, class_332 guiGraphics) {
        if (inspectedBounds != null) {
            class_2561 label = inspectedLabel.method_27661().method_27693(inspectedElement);
            int index = this.inspectedBounds.method_49618() != container.method_49618() ? 0 : 1;
            this.renderInfo(minecraft, guiGraphics, this.inspectedBounds, label, Position.TOP_LEFT, index, true);
        }
    }

    private void renderScreenName(class_310 minecraft, class_437 screen, class_8030 bounds, class_332 guiGraphics) {
        class_2561 label = screenLabel.method_27661().method_27693(getClassName(screen));
        this.renderInfo(minecraft, guiGraphics, bounds, label, Position.BOTTOM_RIGHT, false);
    }

    private void renderVirtualInfo(class_310 minecraft, class_8030 screenBounds, class_332 guiGraphics) {
        class_2561 virtualMode = virtualModeLabel.method_27661().method_10852(CursorsExtended.getInstance().getDisplay().isVirtual() ? class_5244.field_24332 : class_5244.field_24333);
        renderInfo(minecraft, guiGraphics, screenBounds, virtualMode, Position.TOP_RIGHT, false);
    }

    private void renderInfo(class_310 minecraft, class_332 guiGraphics, class_8030 bounds, class_2561 label, Position pos, boolean outline) {
        this.renderInfo(minecraft, guiGraphics, bounds, label, pos, 0, outline);
    }

    private void renderInfo(class_310 minecraft, class_332 guiGraphics, class_8030 bounds, class_2561 label, Position pos, int index, boolean outline) {
        class_5251 textColor = label.method_10866().method_10973();
        int color = 0xFF000000 | (textColor != null ? textColor.method_27716() : 0xFFFFFFFF);

        int textX = pos.getX(minecraft, bounds, label, TEXT_SCALE);
        int textY = pos.getY(minecraft, bounds, index, TEXT_SCALE);

        Matrix3x2fStack matrix3x2fStack = guiGraphics.method_51448().pushMatrix();
        guiGraphics.method_71048();
        if (outline)
            DrawUtil.renderOutline(guiGraphics, bounds.method_49620(), bounds.method_49618(), bounds.comp_1196(), bounds.comp_1197(), color);
        matrix3x2fStack.translate(TEXT_SCALE, TEXT_SCALE);
        matrix3x2fStack.scale(TEXT_SCALE, TEXT_SCALE);
        guiGraphics.method_27535(minecraft.field_1772, label, textX, textY, color);
        matrix3x2fStack.popMatrix();
    }

    private class_8030 getBounds(@Nullable class_364 element) {
        if (element instanceof class_8021 layoutElement) {
            return layoutElement.method_48202();
        }
        return element != null ? element.method_48202() : class_8030.method_48248();
    }

    private String getClassName(class_364 element) {
        String namespace = Services.PLATFORM.isDevelopmentEnvironment() ? "named" : "intermediary";
        return Services.PLATFORM.unmapClassName(namespace, element.getClass().getName());
    }

    private static @Nullable class_364 findDeepest(class_4069 parent, double mouseX, double mouseY) {
        if (CursorTypeUtil.isHovered(parent, mouseX, mouseY)) {
            for (class_364 child : parent.method_25396()) {
                if (child instanceof class_4069 nestedParent) {
                    class_364 nestedChild = findDeepest(nestedParent, mouseX, mouseY);
                    if (nestedChild != null) {
                        return nestedChild;
                    }
                }
                if (CursorTypeUtil.isHovered(child, mouseX, mouseY)) {
                    return child;
                }
            }
        }
        return null;
    }

    public enum Position {
        TOP_LEFT,
        TOP_RIGHT,
        BOTTOM_LEFT,
        BOTTOM_RIGHT;

        private static final int PADDING = 2;

        public int getX(class_310 minecraft, class_8030 bounds, class_2561 text, float scale) {
            int screenWidth = minecraft.method_22683().method_4486();
            int textWidth = minecraft.field_1772.method_27525(text);

            int x = switch (this) {
                case TOP_LEFT, BOTTOM_LEFT -> (int) ((bounds.method_49620() + PADDING) / scale);
                case TOP_RIGHT, BOTTOM_RIGHT -> {
                    int unscaledX = bounds.method_49620() + bounds.comp_1196() - (int) (textWidth * scale) - PADDING;
                    yield (int) (unscaledX / scale);
                }
            };

            if ((x + textWidth) * scale > screenWidth) {
                x = (int) ((screenWidth - textWidth * scale - PADDING) / scale);
                if (x < 0) x = 0;
            }

            return x;
        }

        public int getY(class_310 minecraft, class_8030 bounds, int index, float scale) {
            int screenHeight = minecraft.method_22683().method_4502();
            int textHeight = minecraft.field_1772.field_2000;
            int offsetY = index * textHeight;

            return switch (this) {
                case TOP_LEFT, TOP_RIGHT -> (int) ((bounds.method_49618() + PADDING + offsetY) / scale);
                case BOTTOM_LEFT, BOTTOM_RIGHT -> {
                    if (bounds.comp_1197() == 0 && bounds.comp_1196() == 0) {
                        yield (int) ((screenHeight - textHeight - PADDING - offsetY) / scale);
                    }
                    yield (int) ((bounds.method_49618() - PADDING + Math.max(0, bounds.comp_1197() - textHeight) + offsetY) / scale);
                }
            };
        }
    }
}

==> cursors_extended-fabric-4.1.4%2B1.21.11-sources.jar: io/github/fishstiz/cursors_extended/cursor/debug/CursorDebugRendererImpl.java <==
package io.github.fishstiz.cursors_extended.cursor.debug;

import io.github.fishstiz.cursors_extended.CursorsExtended;
import io.github.fishstiz.cursors_extended.platform.Services;
import io.github.fishstiz.cursors_extended.util.CursorTypeUtil;
import io.github.fishstiz.cursors_extended.util.DrawUtil;
import org.joml.Matrix3x2fStack;
import org.jspecify.annotations.Nullable;

import java.util.function.Supplier;
import net.minecraft.class_124;
import net.minecraft.class_2561;
import net.minecraft.class_310;
import net.minecraft.class_332;
import net.minecraft.class_364;
import net.minecraft.class_4069;
import net.minecraft.class_437;
import net.minecraft.class_5244;
import net.minecraft.class_5251;
import net.minecraft.class_8021;
import net.minecraft.class_8030;

class CursorDebugRendererImpl implements CursorDebugRenderer {
    private static final float TEXT_SCALE = 0.75f;
    private final class_2561 screenLabel = class_2561.method_43470("S: ").method_54663(0xFF339BFF); // blue
    private final class_2561 deepestLabel = class_2561.method_43470("D: ").method_54663(0xFF00FF00); // green
    private final class_2561 inspectedLabel = class_2561.method_43470("I: ").method_54663(0xFFFF0000); // red
    private final class_2561 virtualModeLabel = class_2561.method_43470("Virtual Mode: ").method_27692(class_124.field_1065);
    private class_8030 inspectedBounds;
    private String inspectedElement;

    @Override
    public boolean isActive() {
        return true;
    }

    @Override
    public void setLastCursorAt(class_364 inspected, double mouseX, double mouseY) {
        if (CursorTypeUtil.isHovered(inspected, mouseX, mouseY)) {
            this.inspectedBounds = getBounds(inspected);
            this.inspectedElement = getClassName(inspected);
        }
    }

    @Override
    public void render(class_310 minecraft, Supplier<@Nullable class_437> visibleScreen, class_332 guiGraphics, double mouseX, double mouseY) {
        class_437 screen = visibleScreen.get();
        if (screen != null) {
            class_8030 screenRectangle = getBounds(screen);
            renderScreenName(minecraft, screen, screenRectangle, guiGraphics);
            renderInspected(minecraft, renderDeepest(minecraft, screen, guiGraphics, mouseX, mouseY), guiGraphics);
            renderVirtualInfo(minecraft, screenRectangle, guiGraphics);
        }
    }

    private class_8030 renderDeepest(class_310 minecraft, class_437 screen, class_332 guiGraphics, double mouseX, double mouseY) {
        class_364 child = findDeepest(screen, mouseX, mouseY);
        class_364 inspect = child != null ? child : screen;
        class_8030 bounds = getBounds(inspect);
        class_2561 label = deepestLabel.method_27661().method_27693(getClassName(inspect));
        renderInfo(minecraft, guiGraphics, bounds, label, Position.TOP_LEFT, true);
        return bounds;
    }

    private void renderInspected(class_310 minecraft, class_8030 container, class_332 guiGraphics) {
        if (inspectedBounds != null) {
            class_2561 label = inspectedLabel.method_27661().method_27693(inspectedElement);
            int index = this.inspectedBounds.method_49618() != container.method_49618() ? 0 : 1;
            this.renderInfo(minecraft, guiGraphics, this.inspectedBounds, label, Position.TOP_LEFT, index, true);
        }
    }

    private void renderScreenName(class_310 minecraft, class_437 screen, class_8030 bounds, class_332 guiGraphics) {
        class_2561 label = screenLabel.method_27661().method_27693(getClassName(screen));
        this.renderInfo(minecraft, guiGraphics, bounds, label, Position.BOTTOM_RIGHT, false);
    }

    private void renderVirtualInfo(class_310 minecraft, class_8030 screenBounds, class_332 guiGraphics) {
        class_2561 virtualMode = virtualModeLabel.method_27661().method_10852(CursorsExtended.getInstance().getDisplay().isVirtual() ? class_5244.field_24332 : class_5244.field_24333);
        renderInfo(minecraft, guiGraphics, screenBounds, virtualMode, Position.TOP_RIGHT, false);
    }

    private void renderInfo(class_310 minecraft, class_332 guiGraphics, class_8030 bounds, class_2561 label, Position pos, boolean outline) {
        this.renderInfo(minecraft, guiGraphics, bounds, label, pos, 0, outline);
    }

    private void renderInfo(class_310 minecraft, class_332 guiGraphics, class_8030 bounds, class_2561 label, Position pos, int index, boolean outline) {
        class_5251 textColor = label.method_10866().method_10973();
        int color = 0xFF000000 | (textColor != null ? textColor.method_27716() : 0xFFFFFFFF);

        int textX = pos.getX(minecraft, bounds, label, TEXT_SCALE);
        int textY = pos.getY(minecraft, bounds, index, TEXT_SCALE);

        Matrix3x2fStack matrix3x2fStack = guiGraphics.method_51448().pushMatrix();
        guiGraphics.method_71048();
        if (outline)
            DrawUtil.renderOutline(guiGraphics, bounds.method_49620(), bounds.method_49618(), bounds.comp_1196(), bounds.comp_1197(), color);
        matrix3x2fStack.translate(TEXT_SCALE, TEXT_SCALE);
        matrix3x2fStack.scale(TEXT_SCALE, TEXT_SCALE);
        guiGraphics.method_27535(minecraft.field_1772, label, textX, textY, color);
        matrix3x2fStack.popMatrix();
    }

    private class_8030 getBounds(@Nullable class_364 element) {
        if (element instanceof class_8021 layoutElement) {
            return layoutElement.method_48202();
        }
        return element != null ? element.method_48202() : class_8030.method_48248();
    }

    private String getClassName(class_364 element) {
        String namespace = Services.PLATFORM.isDevelopmentEnvironment() ? "named" : "intermediary";
        return Services.PLATFORM.unmapClassName(namespace, element.getClass().getName());
    }

    private static @Nullable class_364 findDeepest(class_4069 parent, double mouseX, double mouseY) {
        if (CursorTypeUtil.isHovered(parent, mouseX, mouseY)) {
            for (class_364 child : parent.method_25396()) {
                if (child instanceof class_4069 nestedParent) {
                    class_364 nestedChild = findDeepest(nestedParent, mouseX, mouseY);
                    if (nestedChild != null) {
                        return nestedChild;
                    }
                }
                if (CursorTypeUtil.isHovered(child, mouseX, mouseY)) {
                    return child;
                }
            }
        }
        return null;
    }

    public enum Position {
        TOP_LEFT,
        TOP_RIGHT,
        BOTTOM_LEFT,
        BOTTOM_RIGHT;

        private static final int PADDING = 2;

        public int getX(class_310 minecraft, class_8030 bounds, class_2561 text, float scale) {
            int screenWidth = minecraft.method_22683().method_4486();
            int textWidth = minecraft.field_1772.method_27525(text);

            int x = switch (this) {
                case TOP_LEFT, BOTTOM_LEFT -> (int) ((bounds.method_49620() + PADDING) / scale);
                case TOP_RIGHT, BOTTOM_RIGHT -> {
                    int unscaledX = bounds.method_49620() + bounds.comp_1196() - (int) (textWidth * scale) - PADDING;
                    yield (int) (unscaledX / scale);
                }
            };

            if ((x + textWidth) * scale > screenWidth) {
                x = (int) ((screenWidth - textWidth * scale - PADDING) / scale);
                if (x < 0) x = 0;
            }

            return x;
        }

        public int getY(class_310 minecraft, class_8030 bounds, int index, float scale) {
            int screenHeight = minecraft.method_22683().method_4502();
            int textHeight = minecraft.field_1772.field_2000;
            int offsetY = index * textHeight;

            return switch (this) {
                case TOP_LEFT, TOP_RIGHT -> (int) ((bounds.method_49618() + PADDING + offsetY) / scale);
                case BOTTOM_LEFT, BOTTOM_RIGHT -> {
                    if (bounds.comp_1197() == 0 && bounds.comp_1196() == 0) {
                        yield (int) ((screenHeight - textHeight - PADDING - offsetY) / scale);
                    }
                    yield (int) ((bounds.method_49618() - PADDING + Math.max(0, bounds.comp_1197() - textHeight) + offsetY) / scale);
                }
            };
        }
    }
}
